_builtin_zip = zip
_builtin_map = map
_builtin_filter = filter


def zip_(*structures, flatten=False):
  return map(
      lambda *x: x if len(x) > 1 else x[0],
      *structures,
      flatten=flatten)


def map_(function, *structures, flatten=False):
  def impl(function, *structures):
    if len(structures) == 0:
      return structures
    if all(isinstance(s, (tuple, list)) for s in structures):
      if len(set(len(x) for x in structures)) > 1:
        raise ValueError('Cannot merge tuples or lists of different length.')
      args = tuple((impl(function, *x) for x in _builtin_zip(*structures)))
      if hasattr(structures[0], '_fields'):  # namedtuple
        return type(structures[0])(*args)
      else:  # tuple, list
        return type(structures[0])(args)
    if all(isinstance(s, dict) for s in structures):
      if len(set(frozenset(x.keys()) for x in structures)) > 1:
        raise ValueError('Cannot merge dicts with different keys.')
      merged = {
          k: impl(function, *(s[k] for s in structures))
          for k in structures[0]}
      return type(structures[0])(merged)
    return function(*structures)

  result = impl(function, *structures)
  if flatten:
    result = flatten_(result)
  return result


def flatten_(structure):
  if isinstance(structure, dict):
    result = ()
    for key in sorted(list(structure.keys())):
        result += flatten_(structure[key])
    return result
  if isinstance(structure, (tuple, list)):
    result = ()
    for element in structure:
      result += flatten_(element)
    return result
  return (structure,)


def filter_(predicate, *structures, flatten=False):
  def impl(predicate, *structures):
    if len(structures) == 0:
      return structures
    if all(isinstance(s, (tuple, list)) for s in structures):
      if len(set(len(x) for x in structures)) > 1:
        raise ValueError('Cannot merge tuples or lists of different length.')
      # Only wrap in tuples if more than one structure provided.
      if len(structures) > 1:
        filtered = (impl(predicate, *x) for x in _builtin_zip(*structures))
      else:
        filtered = (impl(predicate, x) for x in structures[0])
      # Remove empty containers and construct result structure.
      if hasattr(structures[0], '_fields'):  # namedtuple
        filtered = (x if x != () else None for x in filtered)
        return type(structures[0])(*filtered)
      else:  # tuple, list
        filtered = (
            x for x in filtered if not isinstance(x, (tuple, list, dict)) or x)
        return type(structures[0])(filtered)
    if all(isinstance(s, dict) for s in structures):
      if len(set(frozenset(x.keys()) for x in structures)) > 1:
        raise ValueError('Cannot merge dicts with different keys.')
      # Only wrap in tuples if more than one structure provided.
      if len(structures) > 1:
        filtered = {
            k: impl(predicate, *(s[k] for s in structures))
            for k in structures[0]}
      else:
        filtered = {k: impl(predicate, v) for k, v in structures[0].items()}
      # Remove empty containers and construct result structure.
      filtered = {
          k: v for k, v in filtered.items()
          if not isinstance(v, (tuple, list, dict)) or v}
      return type(structures[0])(filtered)
    if len(structures) > 1:
      return structures if predicate(*structures) else ()
    else:
      return structures[0] if predicate(structures[0]) else ()

  result = impl(predicate, *structures)
  if flatten:
    result = flatten_(result)
  return result


zip = zip_
map = map_
flatten = flatten_
filter = filter_
